import re

import xml.etree.ElementTree as etree


def generate_string(track_details):
    title = track_details.get("title", "") or ""
    uri = track_details.get("uri", "") or ""
    albumArtwork = track_details.get("albumArtwork", "") or ""

    return (
        '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" '
        'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" '
        'xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">'
        '<item id="" parentID="" restricted="True">'
        "<dc:title>{0}</dc:title>"
        '<res protocolInfo="*:*:*:*">{1}</res>'
        "<upnp:albumArtURI>{2}</upnp:albumArtURI>"
        "<upnp:class>object.item.audioItem</upnp:class>"
        "</item>"
        "</DIDL-Lite>".format(title, uri, albumArtwork)
    )


def parse(metadata):
    track_details = {}

    if metadata is None:
        return track_details

    try:
        et = etree.fromstring(metadata).find(
            "DIDL-Lite:item",
            {"DIDL-Lite": "urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"},
        )
    except:
        return track_details

    if et is None:
        return track_details

    track_details["type"] = find_element_value(et, "upnp:class", False)
    track_details["title"] = find_element_value(et, "dc:title", False)
    track_details["uri"] = find_element_value(et, "DIDL-Lite:res", False)
    track_details["artist"] = find_element_value(et, "upnp:artist", True)
    track_details["composer"] = find_element_value(
        et, "upnp:artist[@role='Composer']", True
    )
    track_details["narrator"] = find_element_value(
        et, "upnp:artist[@role='Narrator']", True
    )
    track_details["performer"] = find_element_value(
        et, "upnp:artist[@role='Performer']", True
    )
    track_details["conductor"] = find_element_value(
        et, "upnp:artist[@role='Conductor']", True
    )
    track_details["albumArtist"] = find_element_value(
        et, "upnp:artist[@role='AlbumArtist']", True
    )
    track_details["genre"] = find_element_value(et, "upnp:genre", True)
    track_details["albumGenre"] = find_element_value(et, "upnp:genre", True)
    track_details["albumTitle"] = find_element_value(et, "upnp:album", False)
    track_details["albumArtwork"] = find_element_value(et, "upnp:albumArtURI", False)
    track_details["artwork"] = find_element_value(et, "upnp:artworkURI", False)
    track_details["year"] = parse_int(find_element_value(et, "dc:date", False))
    track_details["disc"] = parse_int(
        find_element_value(et, "upnp:originalDiscNumber", False)
    )
    track_details["discs"] = parse_int(
        find_element_value(et, "upnp:originalDiscCount", False)
    )
    track_details["track"] = parse_int(
        find_element_value(et, "upnp:originalTrackNumber", False)
    )
    track_details["tracks"] = parse_int(
        find_element_value(et, "upnp:originalTrackCount", False)
    )
    track_details["author"] = find_element_value(et, "dc:author", True)
    track_details["publisher"] = find_element_value(et, "dc:publisher", False)
    track_details["published"] = find_element_value(et, "dc:published", False)
    track_details["description"] = find_element_value(et, "dc:description", False)
    track_details["rating"] = find_element_value(et, "upnp:rating", False)
    track_details["channels"] = parse_int(
        find_element_attribute_value(et, "DIDL-Lite:res", "nrAudioChannels")
    )
    track_details["bitDepth"] = parse_int(
        find_element_attribute_value(et, "DIDL-Lite:res", "bitsPerSample")
    )
    track_details["sampleRate"] = parse_int(
        find_element_attribute_value(et, "DIDL-Lite:res", "sampleFrequency")
    )
    track_details["bitRate"] = parse_int(
        find_element_attribute_value(et, "DIDL-Lite:res", "bitrate")
    )
    track_details["duration"] = parse_duration(
        find_element_attribute_value(et, "DIDL-Lite:res", "duration")
    )
    track_details["mimeType"] = find_element_attribute_value(
        et, "DIDL-Lite:res", "protocolInfo"
    )
    #    "provider": "{provider}" // oh:provider
    #    "work": "{work}" // oh:work
    #    "movement": "{movement}" // oh:movement
    #    "show": "{show}" // oh:show
    #    "episode": {episode} // oh:episodeNumber
    #    "episodes": {episodes} // oh:episodeCount
    #    "website": "{website}" // oh:website
    #    "location": "{location}" // oh:location // ISO 6709
    #    "details": "{details}" // oh:details
    #    "extensions": "{extensions}" // oh:extensions // stringified json

    return track_details


def parse_duration(value):
    if value == None:
        return None

    intFinder = re.compile(r"\d+")
    numbers = intFinder.findall(value.split(".")[0])

    if len(numbers) == 3:
        return int(numbers[2]) + (int(numbers[1]) * 60) + (int(numbers[0]) * 360)
    if len(numbers) == 2:
        return int(numbers[1]) + (int(numbers[0]) * 60)
    if len(numbers) == 1:
        return int(numbers[0])

    return None


def parse_int(value):
    if value == None:
        return None

    intFinder = re.compile(r"\d+")
    numbers = intFinder.findall(value)
    if len(numbers) > 0:
        return int(numbers[0])

    return None


def find_element_attribute_value(et, itemKey, itemAttribute):
    namespaces = {
        "DIDL-Lite": "urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/",
        "upnp": "urn:schemas-upnp-org:metadata-1-0/upnp/",
        "dc": "http://purl.org/dc/elements/1.1/",
    }

    value = et.find(itemKey, namespaces)
    parsedValue = None

    if value != None:
        attribute = value.attrib.get(itemAttribute, None)
        if attribute != None:
            parsedValue = attribute

    return parsedValue


def find_element_value(et, itemKey, isArray):
    namespaces = {
        "DIDL-Lite": "urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/",
        "upnp": "urn:schemas-upnp-org:metadata-1-0/upnp/",
        "dc": "http://purl.org/dc/elements/1.1/",
    }

    items = et.findall(itemKey, namespaces)

    parsedValue = None

    if len(items) > 0:
        parsedValue = items[0].text

    if isArray:
        values = set()
        array = []
        if len(items) > 0:
            for i in items:
                value = i.text
                if value and value not in values:
                    values.add(i.text)
                    array.append(i.text)
        return list(array)
    else:
        return parsedValue



import unittest
import os
import asyncio

from openhomedevice.device import Device
from aioresponses import aioresponses


def async_test(coro):
    def wrapper(*args, **kwargs):
        loop = asyncio.new_event_loop()
        try:
            return loop.run_until_complete(coro(*args, **kwargs))
        finally:
            loop.close()

    return wrapper


class FakeAction:
    def __init__(self, response=None):
        self.was_called_times = 0
        self.response = response

    async def async_call(self, **kwargs):
        self.arguments = kwargs
        self.was_called_times += 1
        return self.response


def product_actions():
    return {
        "Product": FakeAction({"Name": b"My Friendly Name", "Room": b"Bathroom"}),
        "SetStandby": FakeAction(),
        "Standby": FakeAction({"Value": True}),
        "SourceIndex": FakeAction({"Value": 3}),
        "Source": FakeAction({"Type": "Playlist", "Name": "Playlist"}),
        "SetSourceIndex": FakeAction(),
        "SourceXml": FakeAction(
            {
                "Value": "<SourceList><Source><Name>Playlist</Name><Type>Playlist</Type><Visible>true</Visible><SystemName>Playlist</SystemName></Source><Source><Name>Radio</Name><Type>Radio</Type><Visible>false</Visible><SystemName>Radio</SystemName></Source><Source><Name>UPnP AV</Name><Type>UpnpAv</Type><Visible>false</Visible><SystemName>UPnP AV</SystemName></Source><Source><Name>Songcast</Name><Type>Receiver</Type><Visible>false</Visible><SystemName>Songcast</SystemName></Source><Source><Name>Net Aux</Name><Type>NetAux</Type><Visible>false</Visible><SystemName>Net Aux</SystemName></Source><Source><Name>Spotify</Name><Type>Spotify</Type><Visible>false</Visible><SystemName>Spotify</SystemName></Source><Source><Name>Roon</Name><Type>Scd</Type><Visible>true</Visible><SystemName>Roon</SystemName></Source><Source><Name>SpeakerTest</Name><Type>Private</Type><Visible>false</Visible><SystemName>SpeakerTest</SystemName></Source></SourceList>"
            }
        ),
    }


def playlist_actions():
    return {
        "TransportState": FakeAction({"Value": "Playing"}),
        "Play": FakeAction(),
        "Pause": FakeAction(),
        "Stop": FakeAction(),
        "Next": FakeAction(),
        "Previous": FakeAction(),
    }


class FakeService:
    def __init__(self, actions):
        self.actions = actions

    def action(self, action_called):
        return self.actions[action_called]


class OpenhomeDevicePlaylistTest(unittest.TestCase):
    @async_test
    @aioresponses()
    async def setUp(self, mocked):
        LOCATION = "http://mydevice:12345/desc.xml"
        with open(
            os.path.join(os.path.dirname(__file__), "data/softwaredescription.xml")
        ) as file:
            mocked.get(LOCATION, body=file.read())
            mocked.get(
                "http://mydevice:12345/softplayer.local/Upnp/av.openhome.org-Product-2/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/softplayer.local/Upnp/av.openhome.org-Volume-2/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/softplayer.local/Upnp/av.openhome.org-Credentials-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/softplayer.local/Upnp/av.openhome.org-Time-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/softplayer.local/Upnp/av.openhome.org-Info-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/softplayer.local/Upnp/av.openhome.org-Config-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/softplayer.local/Upnp/av.openhome.org-Playlist-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/softplayer.local/Upnp/av.openhome.org-Receiver-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/softplayer.local/Upnp/av.openhome.org-Sender-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/softplayer.local/Upnp/av.openhome.org-Radio-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
        self.sut = Device(LOCATION)
        await self.sut.init()
        soap_request_calls = []
        return super().setUp()

    def test_device_parses_uuid(self):
        self.assertEqual(self.sut.uuid(), "uuid:softplayer.local")

    @async_test
    async def test_transport_state(self):
        self.sut.product_service = FakeService(product_actions())
        self.sut.playlist_service = FakeService(playlist_actions())
        self.assertEqual(await self.sut.transport_state(), "Playing")

    @async_test
    async def test_play(self):
        self.sut.product_service = FakeService(product_actions())
        self.sut.playlist_service = FakeService(playlist_actions())
        await self.sut.play()
        self.assertEqual(self.sut.playlist_service.actions["Play"].was_called_times, 1)

    @async_test
    async def test_stop(self):
        self.sut.product_service = FakeService(product_actions())
        self.sut.playlist_service = FakeService(playlist_actions())
        await self.sut.stop()
        self.assertEqual(self.sut.playlist_service.actions["Stop"].was_called_times, 1)

    @async_test
    async def test_pause(self):
        self.sut.product_service = FakeService(product_actions())
        self.sut.playlist_service = FakeService(playlist_actions())
        await self.sut.pause()
        self.assertEqual(self.sut.playlist_service.actions["Pause"].was_called_times, 1)

    @async_test
    async def test_skip_forward(self):
        self.sut.product_service = FakeService(product_actions())
        self.sut.playlist_service = FakeService(playlist_actions())
        await self.sut.skip(10)
        self.assertEqual(self.sut.playlist_service.actions["Next"].was_called_times, 10)

    @async_test
    async def test_skip_backwards(self):
        self.sut.product_service = FakeService(product_actions())
        self.sut.playlist_service = FakeService(playlist_actions())
        await self.sut.skip(-10)
        self.assertEqual(
            self.sut.playlist_service.actions["Previous"].was_called_times, 10
        )

    @async_test
    async def test_skip_nowhere(self):
        self.sut.product_service = FakeService(product_actions())
        self.sut.playlist_service = FakeService(playlist_actions())
        await self.sut.skip(0)
        self.assertEqual(
            self.sut.playlist_service.actions["Previous"].was_called_times, 0
        )
        self.assertEqual(self.sut.playlist_service.actions["Next"].was_called_times, 0)



import unittest
import os
import asyncio

from openhomedevice.device import Device
from aioresponses import aioresponses
from openhomedevice.didl_lite import generate_string

def escape(str):
    str = str.replace("&", "&amp;")
    return str

def async_test(coro):
    def wrapper(*args, **kwargs):
        loop = asyncio.new_event_loop()
        try:
            return loop.run_until_complete(coro(*args, **kwargs))
        finally:
            loop.close()

    return wrapper


class FakeAction:
    def __init__(self, response=None):
        self.was_called_times = 0
        self.response = response

    async def async_call(self, **kwargs):
        self.arguments = kwargs
        self.was_called_times += 1
        return self.response


def pins_actions():
    return {
        "GetDeviceMax": FakeAction({"DeviceMax": 6}),
        "GetIdArray": FakeAction(
            {"IdArray": "[1,2,0,3,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]"}
        ),
        "ReadList": FakeAction(
            {
                "List": '[{"id":1,"mode":"transport","type":"source","uri":"transport:\\/\\/source?udn=4c494e4e-0026-0f22-2963-01387403013f&id=HDMI2&version=1","title":"Sky","description":"","artworkUri":"external:///source?type=Hdmi&systemName=HDMI2","shuffle":false},{"id":2,"mode":"transport","type":"source","uri":"transport:\\/\\/source?udn=4c494e4e-0026-0f22-2963-01387403013f&id=HDMI1&version=1","title":"Playstation 4","description":"","artworkUri":"external:///source?type=Hdmi&systemName=HDMI1","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":3,"mode":"transport","type":"source","uri":"transport:\\/\\/source?udn=4c494e4e-0026-0f22-2963-01387403013f&id=HDMI3&version=1","title":"Fire Stick","description":"","artworkUri":"external:///source?type=Hdmi&systemName=HDMI3","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":4,"mode":"transport","type":"source","uri":"transport:\\/\\/source?udn=4c494e4e-0026-0f22-2963-01387403013f&id=Balanced&version=1","title":"LP12","description":"","artworkUri":"external:///source?type=Analog&systemName=Balanced","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false},{"id":0,"mode":"","type":"","uri":"","title":"","description":"","artworkUri":"","shuffle":false}]'
            }
        ),
        "InvokeIndex": FakeAction(),
    }

def firmware_actions():
    return {
        "GetSoftwareStatus": FakeAction({ "SoftwareStatus": '{"status": "on_latest", "current_software": {"version": "4.100.502", "topic": "main", "channel": "release"}}'}),
        "CheckNow": FakeAction(),
        "Apply": FakeAction()
    }


def info_actions():
    return {
        "Track": FakeAction(
            {
                "Uri": "uri=scd://192.168.1.38:38921",
                "Metadata": escape('<?xml version="1.0" encoding="UTF-8"?><DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"><item parentID="-1" restricted="1"><upnp:albumArtURI xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">http://192.168.1.38:9100/api/image/cc7cbb0b22dd07348d535f1c0db8278a?scale=fit&width=512&height=512&format=image%2Fpng</upnp:albumArtURI><upnp:artist xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Beastie Boys</upnp:artist><upnp:artist xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Beastie Boys</upnp:artist><upnp:artist role="AlbumArtist" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Beastie Boys</upnp:artist><upnp:genre xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Alternative Pop/Rock</upnp:genre><upnp:genre xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Alternative Rap</upnp:genre><upnp:genre xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Alternative/Indie Rock</upnp:genre><upnp:genre xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Pop/Rock</upnp:genre><upnp:genre xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Rap</upnp:genre><upnp:album xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Ill Communication</upnp:album><upnp:artist role="composer" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Adam Yauch</upnp:artist><upnp:artist role="composer" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Mike D</upnp:artist><upnp:artist role="composer" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Ad-Rock</upnp:artist><dc:date xmlns:dc="http://purl.org/dc/elements/1.1/">2009</dc:date><dc:title xmlns:dc="http://purl.org/dc/elements/1.1/">Sabotage</dc:title><oh:originalDiscNumber xmlns:oh="http://www.openhome.org">1</oh:originalDiscNumber><upnp:originalTrackNumber xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">6</upnp:originalTrackNumber><res duration="0:02:58.000">uri=scd://192.168.1.38:38921</res><upnp:class xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">track</upnp:class></item></DIDL-Lite>'),
            }
        )
    }


def volume_actions():
    return {
        "Volume": FakeAction({"Value": 11}),
        "Mute": FakeAction({"Value": True}),
        "SetVolume": FakeAction(),
        "VolumeInc": FakeAction(),
        "VolumeDec": FakeAction(),
        "SetMute": FakeAction(),
    }


def radio_actions():
    return {"SetChannel": FakeAction(), "Play": FakeAction()}


def transport_actions():
    return {
        "TransportState": FakeAction({"State": "Playing"}),
        "Play": FakeAction(),
        "Pause": FakeAction(),
        "Stop": FakeAction(),
        "SkipNext": FakeAction(),
        "SkipPrevious": FakeAction(),
    }


def product_actions():
    return {
        "Product": FakeAction({"Name": b"My Friendly Name", "Room": b"Bathroom"}),
        "SetStandby": FakeAction(),
        "Standby": FakeAction({"Value": True}),
        "SourceIndex": FakeAction({"Value": 3}),
        "Source": FakeAction({"Type": "Analog", "Name": "Front Aux"}),
        "SetSourceIndex": FakeAction(),
        "SourceXml": FakeAction(
            {
                "Value": "<SourceList><Source><Name>Playlist</Name><Type>Playlist</Type><Visible>true</Visible><SystemName>Playlist</SystemName></Source><Source><Name>Radio</Name><Type>Radio</Type><Visible>false</Visible><SystemName>Radio</SystemName></Source><Source><Name>UPnP AV</Name><Type>UpnpAv</Type><Visible>false</Visible><SystemName>UPnP AV</SystemName></Source><Source><Name>Songcast</Name><Type>Receiver</Type><Visible>false</Visible><SystemName>Songcast</SystemName></Source><Source><Name>Net Aux</Name><Type>NetAux</Type><Visible>false</Visible><SystemName>Net Aux</SystemName></Source><Source><Name>Spotify</Name><Type>Spotify</Type><Visible>false</Visible><SystemName>Spotify</SystemName></Source><Source><Name>Roon</Name><Type>Scd</Type><Visible>true</Visible><SystemName>Roon</SystemName></Source><Source><Name>SpeakerTest</Name><Type>Private</Type><Visible>false</Visible><SystemName>SpeakerTest</SystemName></Source></SourceList>"
            }
        ),
    }


class FakeService:
    def __init__(self, actions):
        self.actions = actions

    def action(self, action_called):
        return self.actions[action_called]


class LinnDeviceTests(unittest.TestCase):
    @async_test
    @aioresponses()
    async def setUp(self, mocked):
        LOCATION = "http://mydevice:12345/desc.xml"
        with open(
            os.path.join(os.path.dirname(__file__), "data/linndescription.xml")
        ) as file:
            mocked.get(LOCATION, body=file.read())
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-ConfigApp-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Product-4/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Volume-4/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Credentials-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-OAuth-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Time-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Info-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Config-2/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Transport-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Pins-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-Diagnostics-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-Volkano-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-Privacy-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Exakt-5/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-Configuration-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-Exakt2-4/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-Update-3/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-Cloud-2/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Playlist-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Radio-2/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Receiver-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Sender-2/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-LipSync-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Debug-2/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
        self.sut = Device(LOCATION)
        await self.sut.init()
        soap_request_calls = []
        return super().setUp()

    def test_device_parses_uuid(self):
        self.assertEqual(self.sut.uuid(), "uuid:4c494e4e-1234-ab12-abcd-01234567819f")

    def test_device_parses_model_name(self):
        self.assertEqual(self.sut.model_name(), "My Model")

    def test_device_parses_manufacturer(self):
        self.assertEqual(self.sut.manufacturer(), "Leading Technology Firm")

    def test_device_parses_friendly_name(self):
        self.assertEqual(self.sut.friendly_name(), "Music Machine")

    @async_test
    async def test_device_name(self):
        self.sut.product_service = FakeService(product_actions())
        self.assertEqual(await self.sut.name(), b"My Friendly Name")

    @async_test
    async def test_device_room(self):
        self.sut.product_service = FakeService(product_actions())
        self.assertEqual(await self.sut.room(), b"Bathroom")

    @async_test
    async def test_set_standby_off(self):
        self.sut.product_service = FakeService(product_actions())
        await self.sut.set_standby(False)
        self.assertFalse(
            self.sut.product_service.actions["SetStandby"].arguments["Value"]
        )

    @async_test
    async def test_set_standby_on(self):
        self.sut.product_service = FakeService(product_actions())
        await self.sut.set_standby(True)
        self.assertTrue(
            self.sut.product_service.actions["SetStandby"].arguments["Value"]
        )

    @async_test
    async def test_is_in_standby(self):
        self.sut.product_service = FakeService(product_actions())
        self.assertEqual(await self.sut.is_in_standby(), True)

    @async_test
    async def test_transport_state(self):
        self.sut.transport_service = FakeService(transport_actions())
        self.assertEqual(await self.sut.transport_state(), "Playing")

    @async_test
    async def test_play(self):
        self.sut.transport_service = FakeService(transport_actions())
        await self.sut.play()
        self.assertEqual(self.sut.transport_service.actions["Play"].was_called_times, 1)

    @async_test
    async def test_play_media_with_nothing(self):
        self.sut.radio_service = FakeService(radio_actions())
        await self.sut.play_media(None)
        self.assertEqual(
            self.sut.radio_service.actions["SetChannel"].was_called_times, 0
        )
        self.assertEqual(self.sut.radio_service.actions["Play"].was_called_times, 0)

    @async_test
    async def test_play_media_invalid_url(self):
        self.sut.radio_service = FakeService(radio_actions())
        track_details = {"title": "TITLE", "albumArtwork": "https://host/uri.jpg"}
        await self.sut.play_media(track_details)
        self.assertEqual(
            self.sut.radio_service.actions["SetChannel"].arguments["Uri"], ""
        )
        self.assertEqual(
            self.sut.radio_service.actions["SetChannel"].arguments["Metadata"],
            generate_string(track_details),
        )
        self.assertEqual(self.sut.radio_service.actions["Play"].was_called_times, 1)

    @async_test
    async def test_play_media(self):
        self.sut.radio_service = FakeService(radio_actions())
        track_details = {
            "uri": "https://host/uri.flac",
            "title": "TITLE",
            "albumArtwork": "https://host/uri.jpg",
        }
        await self.sut.play_media(track_details)
        self.assertEqual(
            self.sut.radio_service.actions["SetChannel"].arguments["Uri"],
            "https://host/uri.flac",
        )
        self.assertEqual(
            self.sut.radio_service.actions["SetChannel"].arguments["Metadata"],
            generate_string(track_details),
        )
        self.assertEqual(self.sut.radio_service.actions["Play"].was_called_times, 1)

    @async_test
    async def test_stop(self):
        self.sut.transport_service = FakeService(transport_actions())
        await self.sut.stop()
        self.assertEqual(self.sut.transport_service.actions["Stop"].was_called_times, 1)

    @async_test
    async def test_pause(self):
        self.sut.transport_service = FakeService(transport_actions())
        await self.sut.pause()
        self.assertEqual(
            self.sut.transport_service.actions["Pause"].was_called_times, 1
        )

    @async_test
    async def test_skip_forward(self):
        self.sut.transport_service = FakeService(transport_actions())
        await self.sut.skip(10)
        self.assertEqual(
            self.sut.transport_service.actions["SkipNext"].was_called_times, 10
        )

    @async_test
    async def test_skip_backwards(self):
        self.sut.transport_service = FakeService(transport_actions())
        await self.sut.skip(-10)
        self.assertEqual(
            self.sut.transport_service.actions["SkipPrevious"].was_called_times, 10
        )

    @async_test
    async def test_skip_nowhere(self):
        self.sut.transport_service = FakeService(transport_actions())
        await self.sut.skip(0)
        self.assertEqual(
            self.sut.transport_service.actions["SkipPrevious"].was_called_times, 0
        )
        self.assertEqual(
            self.sut.transport_service.actions["SkipNext"].was_called_times, 0
        )

    @async_test
    async def test_source(self):
        self.sut.product_service = FakeService(product_actions())
        self.assertEqual(
            await self.sut.source(), {"type": "Analog", "name": "Front Aux"}
        )
        self.assertEqual(
            self.sut.product_service.actions["Source"].arguments["Index"], 3
        )

    def test_volume_enabled(self):
        self.assertTrue(self.sut.volume_enabled)

    @async_test
    async def test_volume_level(self):
        self.sut.volume_service = FakeService(volume_actions())
        self.assertEqual(await self.sut.volume(), 11)

    @async_test
    async def test_volume_muted(self):
        self.sut.volume_service = FakeService(volume_actions())
        self.assertTrue(await self.sut.is_muted())

    @async_test
    async def test_set_volume(self):
        self.sut.volume_service = FakeService(volume_actions())
        await self.sut.set_volume(11)
        self.assertEqual(
            self.sut.volume_service.actions["SetVolume"].arguments["Value"], 11
        )

    @async_test
    async def test_increase_volume(self):
        self.sut.volume_service = FakeService(volume_actions())
        await self.sut.increase_volume()
        self.assertEqual(
            self.sut.volume_service.actions["VolumeInc"].was_called_times, 1
        )

    @async_test
    async def test_decrease_volume(self):
        self.sut.volume_service = FakeService(volume_actions())
        await self.sut.decrease_volume()
        self.assertEqual(
            self.sut.volume_service.actions["VolumeDec"].was_called_times, 1
        )

    @async_test
    async def test_set_mute(self):
        self.sut.volume_service = FakeService(volume_actions())
        await self.sut.set_mute(True)
        self.assertTrue(self.sut.volume_service.actions["SetMute"].arguments["Value"])

    @async_test
    async def test_set_source(self):
        self.sut.product_service = FakeService(product_actions())
        await self.sut.set_source(4)
        self.assertEqual(
            self.sut.product_service.actions["SetSourceIndex"].arguments["Value"], 4
        )

    @async_test
    async def test_sources(self):
        self.sut.product_service = FakeService(product_actions())
        self.assertEqual(
            await self.sut.sources(),
            [
                {"index": 0, "name": "Playlist", "type": "Playlist"},
                {"index": 6, "name": "Roon", "type": "Scd"},
            ],
        )

    @async_test
    async def test_track_info(self):
        self.sut.info_service = FakeService(info_actions())
        self.assertEqual(
            await self.sut.track_info(),
            {
                "type": "track",
                "title": "Sabotage",
                "uri": "uri=scd://192.168.1.38:38921",
                "artist": ["Beastie Boys", "Adam Yauch", "Mike D", "Ad-Rock"],
                "composer": [],
                "narrator": [],
                "performer": [],
                "conductor": [],
                "albumArtist": ["Beastie Boys"],
                "genre": [
                    "Alternative Pop/Rock",
                    "Alternative Rap",
                    "Alternative/Indie Rock",
                    "Pop/Rock",
                    "Rap",
                ],
                "albumGenre": [
                    "Alternative Pop/Rock",
                    "Alternative Rap",
                    "Alternative/Indie Rock",
                    "Pop/Rock",
                    "Rap",
                ],
                "albumTitle": "Ill Communication",
                "albumArtwork": "http://192.168.1.38:9100/api/image/cc7cbb0b22dd07348d535f1c0db8278a?scale=fit&width=512&height=512&format=image%2Fpng",
                "artwork": None,
                "year": 2009,
                "disc": None,
                "discs": None,
                "track": 6,
                "tracks": None,
                "author": [],
                "publisher": None,
                "published": None,
                "description": None,
                "rating": None,
                "channels": None,
                "bitDepth": None,
                "sampleRate": None,
                "bitRate": None,
                "duration": 178,
                "mimeType": None,
            },
        )

    @async_test
    async def test_pins(self):
        self.sut.pins_service = FakeService(pins_actions())
        self.assertListEqual(
            await self.sut.pins(),
            [
                {
                    "index": 1,
                    "title": "Sky",
                    "artworkUri": "external:///source?type=Hdmi&systemName=HDMI2",
                },
                {
                    "index": 2,
                    "title": "Playstation 4",
                    "artworkUri": "external:///source?type=Hdmi&systemName=HDMI1",
                },
                {
                    "index": 4,
                    "title": "Fire Stick",
                    "artworkUri": "external:///source?type=Hdmi&systemName=HDMI3",
                },
                {
                    "index": 6,
                    "title": "LP12",
                    "artworkUri": "external:///source?type=Analog&systemName=Balanced",
                },
            ],
        )
        self.assertEqual(
            self.sut.pins_service.actions["ReadList"].arguments["Ids"],
            "[1, 2, 0, 3, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]",
        )

    @async_test
    async def test_invoke_pin(self):
        self.sut.pins_service = FakeService(pins_actions())
        await self.sut.invoke_pin(42)
        self.assertEqual(
            self.sut.pins_service.actions["InvokeIndex"].arguments["Index"], 41
        )
    
    @async_test
    async def test_check_software(self):
        self.sut.update_service = FakeService(firmware_actions())
        result = await self.sut.software_status()
        self.assertEqual(result["status"], "on_latest")
        self.assertEqual(result["current_software"]["version"], "4.100.502")

    @async_test
    async def test_check_latest_firmware(self):
        self.sut.update_service = FakeService(firmware_actions())
        await self.sut.check_latest_firmware()
        self.assertEqual(
            self.sut.update_service.actions["CheckNow"].was_called_times, 1)

    @async_test
    async def test_check_latest_firmware(self):
        self.sut.update_service = FakeService(firmware_actions())
        await self.sut.update_firmware()
        self.assertEqual(
            self.sut.update_service.actions["Apply"].was_called_times, 1)




import json
import asyncio

from async_upnp_client.client_factory import UpnpFactory
from async_upnp_client.aiohttp import AiohttpRequester
# from async_upnp_client.aiohttp import AiohttpNotifyServer

import openhomedevice.didl_lite as didl_lite
import xml.etree.ElementTree as etree


# def on_event(service, service_variables):
#     """Handle a UPnP event."""
#     print(
#         "State variable change for %s, variables: %s",
#         service,
#         ",".join([sv.name for sv in service_variables]),
#     )
#     obj = {
#         "service_id": service.service_id,
#         "service_type": service.service_type,
#         "state_variables": {sv.name: sv.value for sv in service_variables},
#     }
#     print(json.dumps(obj))


class Device(object):
    def __init__(self, location):
        self.location = location

    def setup_services(self):
        self.product_service = self.device.service_id(
            "urn:av-openhome-org:serviceId:Product"
        )
        self.volume_service = self.device.service_id(
            "urn:av-openhome-org:serviceId:Volume"
        )
        self.transport_service = self.device.service_id(
            "urn:av-openhome-org:serviceId:Transport"
        )
        self.playlist_service = self.device.service_id(
            "urn:av-openhome-org:serviceId:Playlist"
        )
        self.info_service = self.device.service_id("urn:av-openhome-org:serviceId:Info")
        self.pins_service = self.device.service_id("urn:av-openhome-org:serviceId:Pins")
        self.radio_service = self.device.service_id(
            "urn:av-openhome-org:serviceId:Radio"
        )
        self.update_service = self.device.service_id("urn:linn-co-uk:serviceId:Update")

    async def init(self):
        requester = AiohttpRequester()
        factory = UpnpFactory(requester)
        self.device = await factory.async_create_device(self.location)
        self.setup_services()

    # async def subscribe(self, service):
    #     service.on_event = on_event
    #     await self.server.event_handler.async_subscribe(service)

    # async def setup_subscriptions(self):
    #     self.server = AiohttpNotifyServer(self.device.requester, 41234)
    #     await self.server.start_server()
    #     print("Listening on: %s", self.server.callback_url)

    #     await self.subscribe(self.product_service)
    #     await self.subscribe(self.volume_service)
    #     await self.subscribe(self.transport_service)
    #     await self.subscribe(self.info_service)

    #     while True:
    #         await asyncio.sleep(120)
    #         await self.server.event_handler.async_resubscribe_all()

    def uuid(self):
        return self.device.udn

    def manufacturer(self):
        return self.device.manufacturer
    
    def model_name(self):
        return self.device.model_name
    
    def friendly_name(self):
        return self.device.friendly_name

    async def name(self):
        action = self.product_service.action("Product")
        return (await action.async_call())["Name"]

    async def room(self):
        action = self.product_service.action("Product")
        return (await action.async_call())["Room"]

    async def set_standby(self, standby_requested):
        await self.product_service.action("SetStandby").async_call(
            Value=standby_requested
        )

    async def is_in_standby(self):
        action = self.product_service.action("Standby")
        return (await action.async_call())["Value"]

    async def transport_state(self):
        if self.transport_service:
            action = self.transport_service.action("TransportState")
            return (await action.async_call()).get("State")

        if (await self.source())["type"] == "Radio":
            action = self.radio_service.action("TransportState")
            return (await action.async_call()).get("Value")

        action = self.playlist_service.action("TransportState")
        return (await action.async_call()).get("Value")

    async def play(self):
        if self.transport_service:
            await self.transport_service.action("Play").async_call()
        else:
            if (await self.source())["type"] == "Radio":
                await self.radio_service.action("Play").async_call()
            else:
                await self.playlist_service.action("Play").async_call()

    async def play_media(self, track_details):
        if self.radio_service and track_details:
            set_channel_action = self.radio_service.action("SetChannel")
            uri = track_details.get("uri", "")
            await set_channel_action.async_call(
                Uri=uri, Metadata=didl_lite.generate_string(track_details)
            )
            await self.radio_service.action("Play").async_call()

    async def stop(self):
        if self.transport_service:
            await self.transport_service.action("Stop").async_call()
        else:
            if (await self.source())["type"] == "Radio":
                await self.radio_service.action("Stop").async_call()
            else:
                await self.playlist_service.action("Stop").async_call()

    async def pause(self):
        if self.transport_service:
            await self.transport_service.action("Pause").async_call()
        else:
            if (await self.source())["type"] == "Radio":
                await self.radio_service.action("Pause").async_call()
            else:
                await self.playlist_service.action("Pause").async_call()

    async def skip(self, offset):
        action = None
        if self.transport_service:
            action = (
                self.transport_service.action("SkipNext")
                if offset > 0
                else self.transport_service.action("SkipPrevious")
            )
        else:
            if (await self.source())["type"] == "Playlist":
                action = (
                    self.playlist_service.action("Next")
                    if offset > 0
                    else self.playlist_service.action("Previous")
                )
        if action:
            for x in range(0, abs(offset)):
                await action.async_call()

    async def source(self):
        index_action = self.product_service.action("SourceIndex")
        source_index = (await index_action.async_call())["Value"]
        source_action = self.product_service.action("Source")
        source_result = await source_action.async_call(Index=source_index)
        return {"type": source_result["Type"], "name": source_result["Name"]}

    @property
    def volume_enabled(self):
        return self.volume_service is not None

    async def volume(self):
        if not self.volume_enabled:
            return None

        action = self.volume_service.action("Volume")
        return (await action.async_call())["Value"]

    async def is_muted(self):
        if not self.volume_enabled:
            return None

        action = self.volume_service.action("Mute")
        result = await action.async_call()
        return result["Value"]

    async def set_volume(self, volume_level):
        if self.volume_enabled:
            action = self.volume_service.action("SetVolume")
            await action.async_call(Value=volume_level)

    async def increase_volume(self):
        if self.volume_enabled:
            await self.volume_service.action("VolumeInc").async_call()

    async def decrease_volume(self):
        if self.volume_enabled:
            await self.volume_service.action("VolumeDec").async_call()

    async def set_mute(self, mute_requested):
        if self.volume_enabled:
            await self.volume_service.action("SetMute").async_call(Value=mute_requested)

    async def set_source(self, index):
        await self.product_service.action("SetSourceIndex").async_call(Value=index)

    async def sources(self):
        action = self.product_service.action("SourceXml")
        result = await action.async_call()
        sources_list_xml = etree.fromstring(result["Value"])

        sources = []
        index = 0
        for source_xml in sources_list_xml:
            visible = source_xml.find("Visible").text == "true"
            if visible:
                sources.append(
                    {
                        "index": index,
                        "name": source_xml.find("Name").text,
                        "type": source_xml.find("Type").text,
                    }
                )
            index = index + 1
        return sources

    async def track_info(self):
        action = self.info_service.action("Track")
        result = await action.async_call()
        return didl_lite.parse(result["Metadata"])

    @property
    def pins_enabled(self):
        return self.device.has_service("urn:av-openhome-org:service:Pins:1")

    async def _get_pin_id_array(self):
        action = self.pins_service.action("GetIdArray")
        result = await action.async_call()
        return json.loads(result["IdArray"])

    async def _pin_metadata(self, ids):
        action = self.pins_service.action("ReadList")
        result = await action.async_call(Ids=json.dumps(ids))
        return json.loads(result["List"])

    async def pins(self):
        if not self.pins_enabled:
            return []

        action = self.pins_service.action("GetDeviceMax")
        max_pins = (await action.async_call())["DeviceMax"]
        pin_id_array = await self._get_pin_id_array()
        pin_metadata = await self._pin_metadata(pin_id_array)

        pins = list()
        for i in range(max_pins):
            if pin_metadata[i].get("id") > 0:
                pin = {
                    "index": i + 1,
                    "title": pin_metadata[i].get("title"),
                    "artworkUri": pin_metadata[i].get("artworkUri"),
                }
                pins.append(pin)
        return pins

    async def invoke_pin(self, pin_id):
        if self.pins_enabled:
            await self.pins_service.action("InvokeIndex").async_call(Index=(pin_id - 1))

    async def software_status(self):
        if self.update_service:
            action = self.update_service.action("GetSoftwareStatus")
            result = await action.async_call()
            return json.loads(result["SoftwareStatus"])

    async def check_latest_firmware(self):
        if self.update_service:
            action = await self.update_service.action("CheckNow").async_call()

    async def update_firmware(self):
        if self.update_service:
            await self.update_service.action("Apply").async_call()




import asyncio
from openhomedevice.device import Device
import time

async def main():
    locations = [
        #"http://192.168.1.32:53101/OsxPlayer-barrys-imac.local/Upnp/device.xml",
        "http://192.168.68.85:55178/4c494e4e-0026-0f21-bf92-01303737013f/Upnp/device.xml"
        # "http://192.168.4x.64:55178/4c494e4e-0026-0f22-2963-01387403013f/Upnp/device.xml"
    ]

    for location in locations:
        device = Device(location)
        await device.init()
        
        # await device.setup_subscriptions()
        print("----")
        print("NAME     : %s" % await device.name())
        print("ROOM     : %s" % await device.room())
        print("UUID     : %s" % device.uuid())
        print("SOURCE   : %s" % await device.source())
        print("STANDBY  : %s" % await device.is_in_standby())
        print("STATE    : %s" % await device.transport_state())
        print("TRACK    : %s" % await device.track_info())
        print("HAS VOL  : %s" % device.volume_enabled)
        print("VOLUME   : %s" % await device.volume())
        print("MUTED    : %s" % await device.is_muted())
        print("SOURCES  : %s" % await device.sources())
        print("HAS PINS : %s" % device.pins_enabled)
        print("PINS     : %s" % await device.pins())
        print("SOFTWARE : %s" % await device.software_status())

        #await device.check_latest_firmware()
        #await device.update_firmware()

        await device.set_standby(False)
        await device.set_volume(30)
        await device.increase_volume()
        await device.decrease_volume()
        await device.set_mute(True)
        await device.set_mute(False)
        await device.play_media({ "uri": "http://opml.radiotime.com/Tune.ashx?id=s50646&formats=mp3,aac,ogg,hls&partnerId=ah2rjr68&username=linnproducts&c=ebrowse"})
        time.sleep(4)
        print("TRACK    : %s" % await device.track_info())
        print("STATE    : %s" % await device.transport_state())
        await device.pause()
        await device.play()
        await device.invoke_pin(1)
        time.sleep(4)
        print("TRACK    : %s" % await device.track_info())
        print("STATE    : %s" % await device.transport_state())
        await device.skip(2)
        time.sleep(0.5)
        print("TRACK    : %s" % await device.track_info())
        print("STATE    : %s" % await device.transport_state())
        await device.stop()
        await device.set_source(0)
        await device.set_source(1)
        print("SOURCE   : %s" % await device.source())

        await device.set_standby(True)
        print("STANDBY  : %s" % await device.is_in_standby())

    print("----")


loop = asyncio.get_event_loop()
loop.run_until_complete(main())



import unittest

from openhomedevice.didl_lite import generate_string, parse, parse_duration, parse_int


class DidlLiteTests(unittest.TestCase):
    def test_int_parsing(self):
        self.assertEqual(parse_duration("42"), 42)
        self.assertEqual(parse_duration("42.5"), 42)
        self.assertIsNone(parse_int("forty"))
        self.assertIsNone(parse_int(None))

    def test_duration_parsing(self):
        self.assertEqual(parse_duration("0:07:40.000"), 460)
        self.assertEqual(parse_duration("1:00.000"), 60)
        self.assertEqual(parse_duration("42.000"), 42)
        self.assertEqual(parse_duration("2:0.5"), 120)
        self.assertIsNone(parse_duration("forty"))
        self.assertIsNone(parse_duration(None))

    def test_parse_empty_didlite(self):
        result = parse(None)
        self.assertEqual(result, {})

    def test_parse_corrupt_didlite(self):
        result = parse(
            '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"><item id="" parentID="" restricted="True"><dc:title></dc:title><res protocolInfo="*:*:*:*"></res><upnp:albumArtURI></upnp:albumArtURI><upnp:class>object.item.audioItem</upnp:class></itemX></DIDL-Lite>'
        )
        self.assertEqual(result, {})

    def test_parse_didlite_missing_item(self):
        result = parse(
            '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"></DIDL-Lite>'
        )
        self.assertEqual(result, {})

    def test_parse_special_chars(self):
        result = parse(
            '<?xml version="1.0" encoding="UTF-8"?><DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:oh="http://www.openhome.org"><item id="0" parentID="0" restricted="1"><upnp:class>object.item.audioItem.musicTrack</upnp:class><dc:title>Apostrophe&apos;</dc:title><upnp:album>Apostrophe(&apos;)</upnp:album><upnp:artist>Frank Zappa</upnp:artist><upnp:albumArtURI>https://i.scdn.co/image/ab67616d0000b27385b05f4bb3c88cf252f96b68</upnp:albumArtURI><res protocolInfo="spotify:*:audio/L16:*" duration="00:05:49.786/1000" bitsPerSample="16" sampleFrequency="44100" nrAudioChannels="2" size="61562336">spotify://connect</res></item></DIDL-Lite>'
        )
        self.assertEqual(result["title"], "Apostrophe'")

    def test_empty_track_details(self):
        track_details = {}
        result = generate_string(track_details)
        self.assertEqual(
            result,
            '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"><item id="" parentID="" restricted="True"><dc:title></dc:title><res protocolInfo="*:*:*:*"></res><upnp:albumArtURI></upnp:albumArtURI><upnp:class>object.item.audioItem</upnp:class></item></DIDL-Lite>',
        )

    def test_track_details_title_is_none(self):
        track_details = {}
        track_details["title"] = None
        result = generate_string(track_details)
        self.assertEqual(
            result,
            '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"><item id="" parentID="" restricted="True"><dc:title></dc:title><res protocolInfo="*:*:*:*"></res><upnp:albumArtURI></upnp:albumArtURI><upnp:class>object.item.audioItem</upnp:class></item></DIDL-Lite>',
        )

    def test_track_details_uri_is_none(self):
        track_details = {}
        track_details["uri"] = None
        result = generate_string(track_details)
        self.assertEqual(
            result,
            '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"><item id="" parentID="" restricted="True"><dc:title></dc:title><res protocolInfo="*:*:*:*"></res><upnp:albumArtURI></upnp:albumArtURI><upnp:class>object.item.audioItem</upnp:class></item></DIDL-Lite>',
        )

    def test_track_details_albumArtwork_is_none(self):
        track_details = {}
        track_details["albumArtwork"] = None
        result = generate_string(track_details)
        self.assertEqual(
            result,
            '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"><item id="" parentID="" restricted="True"><dc:title></dc:title><res protocolInfo="*:*:*:*"></res><upnp:albumArtURI></upnp:albumArtURI><upnp:class>object.item.audioItem</upnp:class></item></DIDL-Lite>',
        )

    def test_track_details(self):
        track_details = {}
        track_details["albumArtwork"] = "ALBUMARTWORK"
        track_details["title"] = "TITLE"
        track_details["uri"] = "URI"
        result = generate_string(track_details)
        self.assertEqual(
            result,
            '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"><item id="" parentID="" restricted="True"><dc:title>TITLE</dc:title><res protocolInfo="*:*:*:*">URI</res><upnp:albumArtURI>ALBUMARTWORK</upnp:albumArtURI><upnp:class>object.item.audioItem</upnp:class></item></DIDL-Lite>',
        )



import unittest
import os
import asyncio

from openhomedevice.device import Device
from aioresponses import aioresponses


def async_test(coro):
    def wrapper(*args, **kwargs):
        loop = asyncio.new_event_loop()
        try:
            return loop.run_until_complete(coro(*args, **kwargs))
        finally:
            loop.close()

    return wrapper


class DeviceWithNoPinsTests(unittest.TestCase):
    @async_test
    @aioresponses()
    async def setUp(self, mocked):
        LOCATION = "http://mydevice:12345/desc.xml"
        with open(
            os.path.join(os.path.dirname(__file__), "data/nopinsdevice.xml")
        ) as file:
            mocked.get(LOCATION, body=file.read())
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-ConfigApp-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Product-3/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Volume-4/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Credentials-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Time-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Info-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Config-2/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Transport-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Pins-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-Update-2/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-Diagnostics-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-Volkano-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-Privacy-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Exakt-4/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-Configuration-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-Exakt2-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-ExaktInputs-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-Cloud-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Playlist-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Radio-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Receiver-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Sender-2/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/linn.co.uk-LipSync-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
            mocked.get(
                "http://mydevice:12345/4c494e4e-1234-ab12-abcd-01234567819f/Upnp/av.openhome.org-Debug-1/service.xml",
                body="<scpd xmlns=\"urn:schemas-upnp-org:service-1-0\"><serviceStateTable/></scpd>",
            )
        self.sut = Device(LOCATION)
        await self.sut.init()
        soap_request_calls = []
        return super().setUp()

    @async_test
    async def test_pins(self):
        result = await self.sut.pins()
        self.assertListEqual(result, list())

    @async_test
    async def test_invoke_pin(self):
        await self.sut.invoke_pin(42)



from openhomedevice.didl_lite import parse
import unittest


class TrackInfoParserTests(unittest.TestCase):
    def setUp(self):
        self.sut = parse

    def test_one(self):
        element_string = '<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:dlna="urn:schemas-dlna-org:metadata-1-0/" xmlns:sec="http://www.sec.co.kr/" xmlns:pv="http://www.pv.com/pvns/"><item id="qobuz/albums/8424562332058/37122539" parentID="qobuz/albums/8424562332058" restricted="1"><upnp:class>object.item.audioItem.musicTrack</upnp:class><dc:title>Violin Sonata No. 2 in A Minor, BWV 1003 (Arr. for Viola da gamba) : Violin Sonata No. 2 in A Minor, BWV 1003 (Arr. for Viola da gamba): II. Fuga</dc:title><dc:creator>Fahmi Alqhai</dc:creator><upnp:artist>Fahmi Alqhai</upnp:artist><upnp:artist role="Performer">Fahmi Alqhai, Performer - Johann Sebastian Bach, Composer</upnp:artist><dc:publisher>Glossa</dc:publisher> <upnp:albumArtURI>http://static.qobuz.com/images/covers/58/20/8424562332058_600.jpg</upnp:albumArtURI><upnp:albumArtURI dlna:profileID="JPEG_TN">http://static.qobuz.com/images/covers/58/20/8424562332058_230.jpg</upnp:albumArtURI><upnp:genre>Klassiek</upnp:genre><dc:date>2017-01-06</dc:date><dc:description><a href="http://static.qobuz.com/goodies/44/000096244.pdf">Digital booklet</a></dc:description><upnp:album>The Bach Album</upnp:album><upnp:originalTrackNumber>2</upnp:originalTrackNumber><ownerUdn>000974e2-681e-1a36-ffff-ffffa38afd93</ownerUdn><res protocolInfo="http-get:*:audio/x-flac:DLNA.ORG_OP=01;DLNA.ORG_FLAGS=01700000000000000000000000000000" bitsPerSample="16" sampleFrequency="44100" nrAudioChannels="2" duration="0:07:40.000">http://192.168.0.110:58050/stream/audio/b362f0f7a1ff33b176bcf2adde75af96.flac</res></item></DIDL-Lite>'
        result = self.sut(element_string)
        self.assertEqual(result.get("type"), "object.item.audioItem.musicTrack")
        self.assertEqual(
            result.get("title"),
            "Violin Sonata No. 2 in A Minor, BWV 1003 (Arr. for Viola da gamba) : Violin Sonata No. 2 in A Minor, BWV 1003 (Arr. for Viola da gamba): II. Fuga",
        )
        self.assertEqual(
            result.get("uri"),
            "http://192.168.0.110:58050/stream/audio/b362f0f7a1ff33b176bcf2adde75af96.flac",
        )
        self.assertSetEqual(
            set(result.get("artist")),
            set(
                [
                    "Fahmi Alqhai, Performer - Johann Sebastian Bach, Composer",
                    "Fahmi Alqhai",
                ]
            ),
        )
        self.assertSetEqual(set(result.get("conductor")), set())
        self.assertSetEqual(set(result.get("albumArtist")), set())
        self.assertSetEqual(set(result.get("genre")), set(["Klassiek"]))
        self.assertSetEqual(set(result.get("albumGenre")), set(["Klassiek"]))
        self.assertEqual(result.get("albumTitle"), "The Bach Album")
        self.assertEqual(
            result.get("albumArtwork"),
            "http://static.qobuz.com/images/covers/58/20/8424562332058_600.jpg",
        )
        self.assertEqual(result.get("artwork"), None)
        self.assertEqual(result.get("year"), 2017)
        self.assertEqual(result.get("disc"), None)
        self.assertEqual(result.get("discs"), None)
        self.assertEqual(result.get("track"), 2)
        self.assertEqual(result.get("tracks"), None)
        self.assertSetEqual(set(result.get("author")), set())
        self.assertEqual(result.get("publisher"), "Glossa")
        self.assertEqual(result.get("published"), None)
        self.assertEqual(result.get("description"), None)
        self.assertEqual(result.get("rating"), None)
        self.assertEqual(result.get("channels"), 2)
        self.assertEqual(result.get("bitDepth"), 16)
        self.assertEqual(result.get("sampleRate"), 44100)
        self.assertEqual(result.get("bitRate"), None)
        self.assertEqual(result.get("duration"), 460)
        self.assertEqual(
            result.get("mimeType"),
            "http-get:*:audio/x-flac:DLNA.ORG_OP=01;DLNA.ORG_FLAGS=01700000000000000000000000000000",
        )

    def test_two(self):
        element_string = '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"><item id="" parentID="" restricted="True"><dc:title>BBC Radio 6 Music (AAA)</dc:title><res protocolInfo="*:*:*:*" bitrate="40000">http://opml.radiotime.com/Tune.ashx?id=s44491&amp;formats=mp3,wma,aac,ogg,hls&amp;partnerId=ah2rjr68&amp;username=bazwilliams&amp;c=ebrowse</res><upnp:albumArtURI>http://cdn-radiotime-logos.tunein.com/s44491q.png</upnp:albumArtURI><upnp:class>object.item.audioItem</upnp:class></item></DIDL-Lite>'
        result = self.sut(element_string)
        self.assertEqual(result.get("type"), "object.item.audioItem")
        self.assertEqual(result.get("title"), "BBC Radio 6 Music (AAA)")
        self.assertEqual(
            result.get("uri"),
            "http://opml.radiotime.com/Tune.ashx?id=s44491&formats=mp3,wma,aac,ogg,hls&partnerId=ah2rjr68&username=bazwilliams&c=ebrowse",
        )
        self.assertSetEqual(set(result.get("artist")), set())
        self.assertSetEqual(set(result.get("conductor")), set())
        self.assertSetEqual(set(result.get("albumArtist")), set())
        self.assertSetEqual(set(result.get("genre")), set())
        self.assertSetEqual(set(result.get("albumGenre")), set())
        self.assertEqual(result.get("albumTitle"), None)
        self.assertEqual(
            result.get("albumArtwork"),
            "http://cdn-radiotime-logos.tunein.com/s44491q.png",
        )
        self.assertEqual(result.get("artwork"), None)
        self.assertEqual(result.get("year"), None)
        self.assertEqual(result.get("disc"), None)
        self.assertEqual(result.get("discs"), None)
        self.assertEqual(result.get("track"), None)
        self.assertEqual(result.get("tracks"), None)
        self.assertSetEqual(set(result.get("author")), set())
        self.assertEqual(result.get("publisher"), None)
        self.assertEqual(result.get("published"), None)
        self.assertEqual(result.get("description"), None)
        self.assertEqual(result.get("rating"), None)
        self.assertEqual(result.get("channels"), None)
        self.assertEqual(result.get("bitDepth"), None)
        self.assertEqual(result.get("sampleRate"), None)
        self.assertEqual(result.get("bitRate"), 40000)
        self.assertEqual(result.get("duration"), None)
        self.assertEqual(result.get("mimeType"), "*:*:*:*")

    def test_three(self):
        element_string = '<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"><item id="d5112915403826242371-co533" parentID="co533" restricted="0"><dc:title xmlns:dc="http://purl.org/dc/elements/1.1/">The Mystic Trumpeter, Op. 18 - Colin Matthews &amp; Imogen Holst</dc:title><upnp:class xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">object.item.audioItem.musicTrack</upnp:class><upnp:albumArtURI xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">http://192.168.1.126:9790/minimserver/*/music/Albums/RSNO*20-*20Holst*20-*20The*20Planets,*20The*20Mystic*20Trumpeter/09*20The*20Mystic*20Trumpeter,*20Op.*2018*20-*20Colin*20Matthews*20*26*20Imogen*20Holst.mp3/$!picture-633-144623.jpg</upnp:albumArtURI><upnp:album xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Holst: The Planets, The Mystic Trumpeter</upnp:album><upnp:artist xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Royal Scottish National Orchestra; David Lloyd-Jones</upnp:artist><upnp:artist role="AlbumArtist" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Royal Scottish National Orchestra; David Lloyd-Jones</upnp:artist><dc:date xmlns:dc="http://purl.org/dc/elements/1.1/">2001-01-01</dc:date><upnp:genre xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Classical</upnp:genre><upnp:genre xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">Orchestral</upnp:genre><res sampleFrequency="44100" bitrate="24000">http://192.168.1.126:9790/minimserver/*/music/Albums/RSNO*20-*20Holst*20-*20The*20Planets,*20The*20Mystic*20Trumpeter/09*20The*20Mystic*20Trumpeter,*20Op.*2018*20-*20Colin*20Matthews*20*26*20Imogen*20Holst.mp3</res></item></DIDL-Lite>'
        result = self.sut(element_string)
        self.assertEqual(result.get("type"), "object.item.audioItem.musicTrack")
        self.assertEqual(
            result.get("title"),
            "The Mystic Trumpeter, Op. 18 - Colin Matthews & Imogen Holst",
        )
        self.assertEqual(
            result.get("uri"),
            "http://192.168.1.126:9790/minimserver/*/music/Albums/RSNO*20-*20Holst*20-*20The*20Planets,*20The*20Mystic*20Trumpeter/09*20The*20Mystic*20Trumpeter,*20Op.*2018*20-*20Colin*20Matthews*20*26*20Imogen*20Holst.mp3",
        )
        self.assertSetEqual(
            set(result.get("artist")),
            set(["Royal Scottish National Orchestra; David Lloyd-Jones"]),
        )
        self.assertSetEqual(set(result.get("conductor")), set())
        self.assertSetEqual(
            set(result.get("albumArtist")),
            set(["Royal Scottish National Orchestra; David Lloyd-Jones"]),
        )
        self.assertSetEqual(set(result.get("genre")), set(["Orchestral", "Classical"]))
        self.assertSetEqual(
            set(result.get("albumGenre")), set(["Orchestral", "Classical"])
        )
        self.assertEqual(
            result.get("albumTitle"), "Holst: The Planets, The Mystic Trumpeter"
        )
        self.assertEqual(
            result.get("albumArtwork"),
            "http://192.168.1.126:9790/minimserver/*/music/Albums/RSNO*20-*20Holst*20-*20The*20Planets,*20The*20Mystic*20Trumpeter/09*20The*20Mystic*20Trumpeter,*20Op.*2018*20-*20Colin*20Matthews*20*26*20Imogen*20Holst.mp3/$!picture-633-144623.jpg",
        )
        self.assertEqual(result.get("artwork"), None)
        self.assertEqual(result.get("year"), 2001)
        self.assertEqual(result.get("disc"), None)
        self.assertEqual(result.get("discs"), None)
        self.assertEqual(result.get("track"), None)
        self.assertEqual(result.get("tracks"), None)
        self.assertSetEqual(set(result.get("author")), set())
        self.assertEqual(result.get("publisher"), None)
        self.assertEqual(result.get("published"), None)
        self.assertEqual(result.get("description"), None)
        self.assertEqual(result.get("rating"), None)
        self.assertEqual(result.get("channels"), None)
        self.assertEqual(result.get("bitDepth"), None)
        self.assertEqual(result.get("sampleRate"), 44100)
        self.assertEqual(result.get("bitRate"), 24000)
        self.assertEqual(result.get("duration"), None)
        self.assertEqual(result.get("mimeType"), None)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
  name = 'openhomedevice',
  version = '2.2.1',
  author = 'Barry John Williams',
  author_email = 'barry@bjw.me.uk',
  description='Provides an API for requesting information from an Openhome device',
  long_description=long_description,
  long_description_content_type="text/markdown",
  url='https://github.com/bazwilliams/openhomedevice',
  packages=setuptools.find_packages(),
  download_url = 'https://github.com/bazwilliams/openhomedevice/tarball/2.1',
  keywords = ['upnp', 'dlna', 'openhome', 'linn', 'ds', 'music', 'render', 'async'],
  install_requires = ['async_upnp_client>=0.40', 'lxml>=4.8.0'],
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6'
)
